/* ***********************************************************************
 *
 *  umake.cpp: A tool for creating Makefiles automatically
 *
 *  *** NOTE: if the ./umake command (located in the Ubit base directory)
 *  *** does not work properly, you may have to modify CONFIGURE.cpp
 *  *** (but there should be no need to modify this file, i.e. umake.cpp)
 *
 *  Ubit GUI Toolkit - Version 6.3
 *  (C) 2009 Eric Lecolinet | ENST Paris | http://www.enst.fr/~elc/ubit
 *
 * ***********************************************************************
 * COPYRIGHT NOTICE : 
 * THIS PROGRAM IS DISTRIBUTED WITHOUT ANY WARRANTY AND WITHOUT EVEN THE 
 * IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 
 * YOU CAN REDISTRIBUTE IT AND/OR MODIFY IT UNDER THE TERMS OF THE GNU 
 * GENERAL PUBLIC LICENSE AS PUBLISHED BY THE FREE SOFTWARE FOUNDATION; 
 * EITHER VERSION 2 OF THE LICENSE, OR (AT YOUR OPTION) ANY LATER VERSION.
 * SEE FILES 'COPYRIGHT' AND 'COPYING' FOR MORE DETAILS.
 * ***********************************************************************/

#include <cstring>
#include <fstream>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include "umake.hpp"
using namespace std;

#define PREFIX      "lib"
#define TEST_SOURCE PREFIX"_tst.c"
#define TEST_EXEC   PREFIX"_tst"
#define STAT_FILE   PREFIX"_status"


int main(int argc, char** argv) {
  UMake um(argc, argv);
  return 0;
}

const char* getVar(const char* name, const char* def) {
  const char* val = ::getenv(name);
  if (!val || !*val) val = def;
  return val;
}

void UMake::print(const string& msg) {
  cout << msg << endl;
  cerr << msg << endl;
}

UMake::UMake(int argc, char** argv) {
  if (argc < 3) {
    print("Fatal error: umake requires 2 arguments: the names of config and log files");
    exit(1);
  }
  CONFIG_FILE = argv[1];
  CONFIG_LOG = argv[2];
  
  struct utsname name;
  ::uname(&name);
  SYSNAME = name.sysname;
  
  setProjectOptions();

  print("\n**** umake options ********************************************\n");
  printOptions();
    
  print("\n**** searching libraries **************************************\n");
  setProjectConfig();
  
  {
    CXXFLAGS += " -I" + BUILD_INCLUDE_DIR + " ";
    string libs;
    
    for (Libs::iterator i = liblist.begin(); i != liblist.end(); ++i) {
      if ((*i)->found) {
        if (!(*i)->compile_options.empty()) CXXFLAGS += (*i)->compile_options + " ";
        if (!(*i)->link_options.empty()) libs += (*i)->link_options + " ";
      }
    }
        
    if (LD.empty()) LD = CXX;
    LDLIBS = " -L" + BUILD_LIB_DIR + " " + PROJECT_LIBS + " " + libs + ' ' + LDLIBS;
    
    if (LIB_LD.empty()) LIB_LD = "ar vru";  // cf ranlib
    LIB_LDLIBS  = libs + ' '  + LIB_LDLIBS;

    if (RM.empty()) RM = "rm";
    if (CP.empty()) CP = "cp";
    if (INSTALL.empty()) INSTALL = "install";    
  }
  
  
  print("\n**** generating files *****************************************\n");
  createMakefileDefs();
  createFeaturesHeader();

  print("\n***************************************************************\n");
  finalStatus();
}

//==============================================================================

void UMake::printOptions()       // prints out UMake options
{  
  print("- Project Name: "    + PROJECT_NAME);
  print("- Project Version: " + PROJECT_VERSION);
  print("- Windowing toolkit: " + WIN_TOOLKIT);
  print("- Operating System: "+ SYSNAME);
  print("- C Compiler: "      + CC);
  print("- C++ Compiler: "    + CXX);
}

//==============================================================================

void UMake::finalStatus()      // prints out final status
{
  bool main_libs_found = true, aux_libs_found = true;
  
  for (Libs::iterator i = liblist.begin(); i != liblist.end(); ++i) {
    if (! (*i)->found) {
      if ((*i)->mandatory) main_libs_found = false;
      else aux_libs_found = false;
    }
  }
    
  if (main_libs_found) 
    print("All mandatory libraries have been found\n");
  else {
    print("!!! Some MANDATORY libraries have NOT been found:\nEdit "
          + CONFIG_FILE + " to change library paths then start again "
          "(see more info in " + CONFIG_LOG + ")\n");
    exit(1);
  }
  
  if (aux_libs_found) 
    print("All auxilliary libraries have been found\n");
  else {
    print("!!! Some auxiliary libraries have NOT been found:\n"
          "Some features will be unavailable, if needed, edit "
          + CONFIG_FILE
          + " to change library paths the start again "
          "(see more info in "+ CONFIG_LOG + ")\n");
  }
  
  print("\nType 'make' (or 'gmake') to build the project");
  print("Then type 'make install' if you want to install the library\n");
}

//==============================================================================

void UMake::createMakefileDefs()
{
  print("- creating " + MAKEFILE_DEFS);
  ofstream fout(MAKEFILE_DEFS);
  if (!fout)  {
    print("!!! Can't create: " + MAKEFILE_DEFS +
          "\n!!! fix file names or read/write permissions in "+CONFIG_FILE+" then start again");
    exit(1);
  }
  
  for (Libs::iterator i = liblist.begin(); i != liblist.end(); ++i) {
    if ((*i)->found) fout << "WITH_" << (*i)->name <<"=yes" << endl;
  }
  fout << endl;
  
  fout << "WIN_TOOLKIT=" << WIN_TOOLKIT << endl;
  fout << "SYSNAME=" << SYSNAME << endl; 
  fout << endl;

  fout << "CC=" << CXX << endl;
  fout << "CFLAGS=" << CXXFLAGS << endl;

  fout << "CXX=" << CXX << endl;
  fout << "CXXFLAGS=" << CXXFLAGS << endl;
  fout << endl;

  fout << "LD="  << LD <<endl;
  fout << "LDLIBS=" << LDLIBS <<endl;
  fout << endl;
  
  fout << "LIB_LD=" << LIB_LD << endl;
  fout << "LIB_LDLIBS=" << LIB_LDLIBS << endl;
  fout << endl;  

  fout << "RM=" << RM << endl;
  fout << "CP=" << CP << endl;
  fout << "INSTALL=" << INSTALL << endl;
  fout << endl;  
  
  fout << "INSTALL_INCLUDE_DIR=" << INSTALL_INCLUDE_DIR << endl;
  fout << "INSTALL_LIB_DIR=" << INSTALL_LIB_DIR << endl;
  fout << "INSTALL_BIN_DIR=" << INSTALL_BIN_DIR << endl;
  fout << endl;  
  
  fout.close();
}
  
//==============================================================================

void UMake::createFeaturesHeader()
{  
  //string features_header = BUILD_INCLUDE_DIR + "/" + FEATURES_HEADER; !!! A REVOIR
  print("- creating " + FEATURES_HEADER);
  ofstream fout(FEATURES_HEADER);
  if (!fout)  {
    print("!!! Can't create: " + FEATURES_HEADER +
          "\n!!! fix file names or read/write permissions in "+CONFIG_FILE+" then start again");
    exit(1);
  }
  
  int version_major = 0, version_minor = 0, version_micro = 0; 
  sscanf(PROJECT_VERSION.c_str(), "%d.%d.%d", &version_major, &version_minor, &version_micro);
  
  fout << "#ifndef "<< PROJECT_NAME << "_FEATURES_H" << endl;
  fout << "#define "<< PROJECT_NAME << "_FEATURES_H 1" << endl;
  fout << endl;
  fout << "#define "<< PROJECT_NAME << "_VERSION \""<< PROJECT_VERSION << "\"" << endl;
  fout << "#define "<< PROJECT_NAME << "_VERSION_MAJOR " << version_major <<endl;
  fout << "#define "<< PROJECT_NAME << "_VERSION_MINOR " << version_minor <<endl;
  fout << "#define "<< PROJECT_NAME << "_VERSION_MICRO " << version_micro <<endl;
  fout << endl;
  
  for (Libs::iterator i = liblist.begin(); i != liblist.end(); ++i) {
    if ((*i)->found) 
      fout << "#define "<< PROJECT_NAME << "_WITH_" << (*i)->name <<" 1" << endl;
  }
  fout << endl << "#endif" << endl;
  fout.close();
}

//==============================================================================

bool UMake::needLib(const char* name, Paths& paths, 
                    const char* headers, const char* libs, const char* func) {
  Lib* lib = new Lib(name, true);
  //liblist[lib->name] = lib;
  liblist.push_back(lib);
  return lib->searchLib(*this, paths, headers, libs, func);
}

bool UMake::optLib(const char* name, Paths& paths, 
                   const char* headers, const char* libs, const char* func) {
  Lib* lib = new Lib(name, false);
  //liblist[lib->name] = lib;
  liblist.push_back(lib);
  return lib->searchLib(*this, paths, headers, libs, func);
}

void UMake::remLib(const char* name) {
  Libs::iterator found = liblist.end();
  for (Libs::iterator i = liblist.begin(); i != liblist.end(); ++i) {
    if ((*i)->name == name) found = i;
  }
  if (found != liblist.end()) liblist.erase(found);
}

//==============================================================================
/*
bool equals(const char* s1, const char* s2) {   // case insensitive
  for (int k = 0; ; ++k) {
    if (tolower(s1[k]) != tolower(s2[k])) return false;
    if (s1[k] == 0) return true;
  }
}
*/
//==============================================================================

void Paths::add(const char* inc_path, const char* lib_path) {
  head_paths.push_back(inc_path);
  lib_paths.push_back(lib_path);
}

//==============================================================================

Lib::Lib(const char* _name, bool _mand) 
: mandatory(_mand), found(false), name(_name) {
}

/*
bool Lib::searchHeader(Paths& paths, const char* header) {
   bool header_ok = false;
   struct stat st;
   for (unsigned int k = 0; k < paths.head_paths.size(); ++k) {
   string header_path = paths.head_paths[k];
   header_path += '/';
   header_path += header;    
   cerr << "- looking for: " << header_path << endl;
   if (::lstat(header_path.c_str(), &st) == 0) {
   cerr << "- header found: " << header << endl;
   header_ok = true;
   break;
   }
   }     
   if (!header_ok) cerr << " - header not found: " << header << endl;
 return false;
}
*/
//==============================================================================

bool Lib::searchLib(UMake& um, Paths& paths, const char* headers, const char* libs, const char* func) {
   
  const char* mand_txt = (mandatory ? " [MANDATORY] " : " ");
  um.print("- searching for library " + name + mand_txt + "...");
  cerr << "\n### testing function: " << func << " with libraries: " << libs <<endl << endl;
  
  if (paths.head_paths.size() == 0) {  // this may be a framework
    if (searchLib(um, 0, 0, headers, libs, func)) {
      found = true;
      link_options = libs;
      um.print("  *found*");
      return true;
    }
  }
  
  else {
    for (unsigned int k = 0; k < paths.head_paths.size(); ++k) {
      cerr << "\n### Testing with options: " << k
      << (paths.head_paths[k] ? paths.head_paths[k] : "")
      << " " << (paths.lib_paths[k] ? paths.lib_paths[k] : "")
      << endl;
      if (searchLib(um, paths.head_paths[k], paths.lib_paths[k], headers, libs, func)) {
        found = true;
        compile_options = paths.head_paths[k];
        link_options = paths.lib_paths[k] + string(" ") + libs;
        um.print("  *found*");
        return true;
      }
    }
  }
  
  um.print("  *NOT FOUND*");
  return false;  // not found
}

//==============================================================================

bool Lib::searchLib(UMake& um, const char* header_path, const char* lib_path,
                    const char* headers, const char* libs, const char* func) {
  ofstream fout(TEST_SOURCE);
  if (!fout)  {
    um.print("!!! Can't create file: " + string(TEST_SOURCE) +
          "\n!!! fix file names or read/write permissions in "+um.CONFIG_FILE+" then start again");
    exit(1);
  }
  
  fout << "#include <unistd.h>" << endl;
  fout << "#include <stdio.h>" << endl;
  fout << "#include <string.h>" << endl;
  if (headers && headers[0]) {
      if (headers[0] == '#') fout << headers << endl;
      else fout << "#include <" << headers << ">" << endl;
      }
  fout << "int main() {" << endl;
  fout << "  void* p = (void*)" << func << ";" << endl;
  fout << "}" << endl;
  fout.close();
  
  string cmd;
  
  cmd += "cat "TEST_SOURCE" > /dev/stderr";
  ::system(cmd.c_str());
  
  cmd = um.CC;
  if (header_path != 0 && header_path[0] != 0) {
    cmd += " ";
    cmd += header_path;
  }
  if (lib_path != 0 && lib_path[0] != 0) {
    cmd += " ";
    cmd += lib_path;
  }
  if (libs) {
    cmd += " ";
    cmd += libs;
  }
  cmd += " ";
  cmd += TEST_SOURCE;
  
  cmd += " -o ";
  cmd += TEST_EXEC;
  
  cmd += " ; echo $? >"STAT_FILE;
  
  cerr << "\n### Compiling with command: " << endl << cmd <<endl;
  system(cmd.c_str());
  
  ifstream fin(STAT_FILE);
  if (!fin)  {
    um.print("!!! Can't read file: " + string(STAT_FILE) +
             "\n!!! fix file names or read/write permissions in "+um.CONFIG_FILE+" then start again");
    exit(1);
  }  
  
  int cmd_stat = -1;
  fin >> cmd_stat;
  cerr << "\n### Compilation status: " << cmd_stat << endl << endl;
  return (cmd_stat == 0);
}
