/* ***********************************************************************
 *
 *  umake.hpp: Ubit tool for confuring makefiles
 *  Ubit GUI Toolkit - Version 6.0
 *  (C) 2008 Eric Lecolinet | ENST Paris | www.enst.fr/~elc/ubit
 *
 * ***********************************************************************
 * COPYRIGHT NOTICE : 
 * THIS PROGRAM IS DISTRIBUTED WITHOUT ANY WARRANTY AND WITHOUT EVEN THE 
 * IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 
 * YOU CAN REDISTRIBUTE IT AND/OR MODIFY IT UNDER THE TERMS OF THE GNU 
 * GENERAL PUBLIC LICENSE AS PUBLISHED BY THE FREE SOFTWARE FOUNDATION; 
 * EITHER VERSION 2 OF THE LICENSE, OR (AT YOUR OPTION) ANY LATER VERSION.
 * SEE FILES 'COPYRIGHT' AND 'COPYING' FOR MORE DETAILS.
 * ***********************************************************************/

#ifndef _umake_hpp_
#define	_umake_hpp_ 1

#include <vector>
#include <list>
#include <string>
#include <iostream>
#include <fstream>

class UMake;
extern const char* getVar(const char* name, const char* def);

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Possible Header and Library Paths.
struct Paths {
  void add(const char* header_path, const char* library_path);
  std::vector <const char*> head_paths;
  std::vector <const char*> lib_paths;
};

/// Specifies a Library that will be used in the compilation process.
struct Lib {
  Lib(const char* name, bool mandatory);
  
  bool searchLib(UMake&, Paths&, const char* headers, const char* libs, const char* func);
  
  bool searchLib(UMake&, const char* header_path, const char* lib_path,
                 const char* headers, const char* libs, const char* func);

  bool mandatory, found;
  std::string name;
  std::string compile_options, link_options;
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// UMake configures the build system accordng to system and client requirements.
class UMake {
public:
  UMake(int argc, char* argv[]);
  
  std::string
  CONFIG_FILE,
  CONFIG_LOG,
  
  PROJECT_NAME,
  PROJECT_VERSION,
  WIN_TOOLKIT,
  PROJECT_LIBS,
  
  SYSNAME,     ///< Operating System name (e.g. Darwin, MacOSX, Linux...).
  CC,          ///< C Compiler.
  CXX,         ///< C++ Compiler.
  CFLAGS,      ///< C Compiler Flags.
  CXXFLAGS,    ///< C++ Compiler Flag.
  LD,          ///< Linker command for creating the Project Executables
  LDLIBS,      ///< Libraries for creating the Project Executables.
  LIB_LD,      ///< Linker command for creating the Project Library.
  LIB_LDLIBS,  ///< Libraries for creating the Project Library.
  RM,
  CP,
  INSTALL,
  
  // generated files
  MAKEFILE_DEFS,
  FEATURES_HEADER,
  
  // compilation directories
  BUILD_INCLUDE_DIR,
  BUILD_LIB_DIR,
  BUILD_BIN_DIR,
  
  // installation directories
  INSTALL_INCLUDE_DIR,
  INSTALL_LIB_DIR,
  INSTALL_BIN_DIR;
  
  bool needLib(const char* lib_name, Paths& lib_paths, 
               const char* headers, const char* libraries, const char* func);
  /**< specifies that this library is needed (and mandatory).
   *
   *  needLib(name, paths, headers, libspec, function)
   *  optLib(name, paths, headers, libspec, function)
   *
   *  These functions search for a library and its header and modify the generated
   *  files as needed (#1). They return true if the Lib is available. needLib() specifies
   *  that the Lib is mandatory to create the project while optLib() indicates an optional
   *   library (some features will not be available by the project can be created)
   *
   *  Arguments:
   *  - 'name'    : the name of the requested library. #define UBIT_WITH_name 1 will be added
   *    to the generated 'features.h' file if this lb is found.
   *  - 'paths'   : the appropriate multipath that was previously defined. A multipath 
   *    specifies a list of possible (header_path,library_path) combinations
   *  - 'headers' : the name of the required header(s)
   *       #- ex with 1 header:  "Xlib.h" means than #include <Xlib.h> is needed
   *       #- ex with 2 headers: "#include<ft2build.h>\n#include FT_FREETYPE_H" 
   *          means that this expression (with 2 headers) must be textually used
   *  - 'libspec' : specifies the required libary(ies)
   *       #- ex: "-lXpm -lX11" (Unix) or "-framework OpenGL" (MacOSX)
   *  - 'function': a C function of the library
   *
   * Generated Files:
   *  - ubit_features.h   (to be completed)
   *  - confmake  (to be completed)
   *
   * Principle:
   * - Using the specified arguments, these functions attempt to compile a C program
   *   that use the specified 'headers' and 'libspec'. The compilation will fail if 
   *   one of them cannot be found. These functions do not attempt to find files by
   *   themselves, but rely on the compiler (as will be done during the actual
   *   building phase of the project)
   */
  
  void print(const std::string& msg);

  bool optLib(const char* lib_name, Paths& lib_paths, 
              const char* headers, const char* libraries, const char* func);
  ///< specifies an optional library; @see needLib().
 
  void remLib(const char* lib_name);
  ///< removes a library from the searched lib list.  
  
private:  
  typedef std::list<Lib*> Libs;
  Libs liblist;
  
  void setProjectOptions();    ///< sets compilation options
  void setProjectConfig();     ///< specifies which libraries are needed
  void printOptions();         ///< prints out UMake pptions.
  void finalStatus();          ///< prints out final status.
  void createMakefileDefs();
  void createFeaturesHeader();
};

#endif
